local _G = _G;
local ABGP = ABGP;
local AceGUI = _G.LibStub("AceGUI-3.0");

local CreateFrame = CreateFrame;
local GetItemInfo = GetItemInfo;
local GetItemIcon = GetItemIcon;
local IsModifiedClick = IsModifiedClick;
local GetItemQualityColor = GetItemQualityColor;
local ResetCursor = ResetCursor;
local MouseIsOver = MouseIsOver;
local CursorUpdate = CursorUpdate;
local LE_ITEM_QUALITY_COMMON = LE_ITEM_QUALITY_COMMON;
local LE_ITEM_QUALITY_ARTIFACT = LE_ITEM_QUALITY_ARTIFACT;
local pairs = pairs;
local ipairs = ipairs;
local floor = floor;
local min = min;
local max = max;
local mod = mod;
local table = table;
local select = select;
local math = math;

local function CreateElement(frame, anchor, template)
    local elt = CreateFrame("Button", nil, frame, template);
    elt:SetHeight(frame:GetHeight());
    elt:EnableMouse(true);
    elt:RegisterForClicks("LeftButtonUp", "RightButtonUp");
    elt:SetHyperlinksEnabled(true);
    elt:SetScript("OnHyperlinkEnter", function(self, itemLink)
        _G.ShowUIPanel(_G.GameTooltip);
        _G.GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT");
        _G.GameTooltip:SetHyperlink(itemLink);
        _G.GameTooltip:Show();
        self:GetParent():RequestHighlight(true);
        self.hasItem = itemLink;
        CursorUpdate(self);
    end);
    elt:SetScript("OnHyperlinkLeave", function(self)
        _G.GameTooltip:Hide();
        self:GetParent():RequestHighlight(false);
        self.hasItem = nil;
        ResetCursor();
    end);
    elt:SetScript("OnHyperlinkClick", function(self, itemLink, text, ...)
        if IsModifiedClick() then
            _G.HandleModifiedItemClick(select(2, GetItemInfo(itemLink)));
        else
            self:GetParent().obj:Fire("OnClick", ...);
        end
    end);
    elt:SetScript("OnEnter", function(self)
        self:GetParent():RequestHighlight(true);
    end);
    elt:SetScript("OnLeave", function(self)
        self:GetParent():RequestHighlight(false);
    end);
    elt:SetScript("OnClick", function(self, ...)
        self:GetParent().obj:Fire("OnClick", ...)
    end);
    elt:SetScript("OnUpdate", function(self)
        if _G.GameTooltip:IsOwned(self) and self.hasItem then
            _G.GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT");
            _G.GameTooltip:SetHyperlink(self.hasItem);
            CursorUpdate(self);
        end
    end);

    if anchor then
        elt:SetPoint("TOPLEFT", anchor, "TOPRIGHT");
    else
        elt:SetPoint("TOPLEFT", frame);
    end
    elt:SetPoint("BOTTOM", frame);

    return elt;
end

local function CreateFontString(frame, y)
    local fontstr = frame:CreateFontString(nil, "BACKGROUND", "GameFontNormal");
    fontstr:SetJustifyH("LEFT");
    if y then
        fontstr:SetPoint("LEFT", frame, 2, y);
        fontstr:SetPoint("RIGHT", frame, -2, y);
    else
        fontstr:SetPoint("LEFT", frame, 2, 1);
        fontstr:SetPoint("RIGHT", frame, -2, 1);
    end
    fontstr:SetWordWrap(false);

    return fontstr;
end

do
    local Type, Version = "ABGP_Player", 1;

    local mainSpecFont = CreateFont("ABGPHighlight");
    mainSpecFont:CopyFontObject(GameFontHighlight);
    mainSpecFont:SetTextColor(unpack(ABGP.ColorTable));

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self:SetHeight(32);

            self.player.text:SetText("");
            self.rank.text:SetText("");
            self.priority.text:SetText("");
            self.equipped.textTop:SetText("");
            self.equipped.textMid:SetText("");
            self.equipped.textBot:SetText("");
            self.requestType.text:SetText("");
            self.notes.text:SetText("");

            self.frame.highlightRequests = 0;
            self.frame:UnlockHighlight();

            self.background:Hide();
        end,

        ["SetData"] = function(self, data)
            self.data = data;

            self.player.text:SetText(ABGP:ColorizeName(data.player or "", data.class));
            self.rank.text:SetText(data.rank or "");
            self.rank.text:SetFontObject((data.preferredGroup and data.group == data.preferredGroup) and "ABGPHighlight" or "GameFontNormal");
            self.ep.text:SetText(data.ep and ("%.3f"):format(data.ep) or "--");
            self.gp.text:SetText(data.gp and ("%.3f"):format(data.gp) or "--");
            self.priority.text:SetText(data.priority and ("%.3f"):format(data.priority) or "--");

            if data.equipped then
                if #data.equipped == 2 then
                    self.equipped.textTop:SetText(data.equipped[1]);
                    self.equipped.textBot:SetText(data.equipped[2]);
                elseif #data.equipped == 1 then
                    self.equipped.textMid:SetText(data.equipped[1]);
                end
            end

            local requestTypes = {
                [ABGP.RequestTypes.MS] = "MS",
                [ABGP.RequestTypes.OS] = "OS",
                [ABGP.RequestTypes.ROLL] = "",
            };
            self.requestType.text:SetText(data.requestType and requestTypes[data.requestType] or "");
            local specialFont = (data.requestType and data.requestType == ABGP.RequestTypes.MS) and "ABGPHighlight" or "GameFontHighlight";
            self.requestType.text:SetFontObject(specialFont);

            self.roll.text:SetText(data.roll or "");
            self.roll.text:SetFontObject(data.currentMaxRoll and "ABGPHighlight" or "GameFontHighlight");

            self.notes.text:SetText(data.notes or "");
        end,

        ["SetWidths"] = function(self, widths)
            self.player:SetWidth(widths[1] or 0);
            self.rank:SetWidth(widths[2] or 0);
            self.ep:SetWidth(widths[3] or 0);
            self.gp:SetWidth(widths[4] or 0);
            self.priority:SetWidth(widths[5] or 0);
            self.equipped:SetWidth(widths[6] or 0);
            self.requestType:SetWidth(widths[7] or 0);
            self.roll:SetWidth(widths[8] or 0);
        end,

        ["SetHeight"] = function(self, height)
            self.frame:SetHeight(height);
        end,

        ["ShowBackground"] = function(self, show)
            self.background[show and "Show" or "Hide"](self.background);
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local frame = CreateFrame("Button");
        frame:SetHeight(32);
        frame:Hide();

        frame.highlightRequests = 0;
        frame.RequestHighlight = function(self, enable)
            self.highlightRequests = self.highlightRequests + (enable and 1 or -1);
            self[self.highlightRequests > 0 and "LockHighlight" or "UnlockHighlight"](self);
        end;

        local highlight = frame:CreateTexture(nil, "HIGHLIGHT");
        highlight:SetTexture("Interface\\HelpFrame\\HelpFrameButton-Highlight");
        highlight:SetAllPoints();
        highlight:SetBlendMode("ADD");
        highlight:SetTexCoord(0, 1, 0, 0.578125);

        local background = frame:CreateTexture(nil, "BACKGROUND");
        background:SetAllPoints();
        background:SetColorTexture(0, 0, 0, 0.5);

        local player = CreateElement(frame);
        player.text = CreateFontString(player);

        local rank = CreateElement(frame, player);
        rank.text = CreateFontString(rank);

        local ep = CreateElement(frame, rank);
        ep.text = CreateFontString(ep);

        local gp = CreateElement(frame, ep);
        gp.text = CreateFontString(gp);

        local priority = CreateElement(frame, gp);
        priority.text = CreateFontString(priority);

        local equipped = CreateElement(frame, priority);
        equipped.textTop = CreateFontString(equipped, 8);
        equipped.textMid = CreateFontString(equipped);
        equipped.textBot = CreateFontString(equipped, -4);

        local requestType = CreateElement(frame, equipped);
        requestType.text = CreateFontString(requestType);
        requestType.text:SetJustifyH("CENTER");

        local roll = CreateElement(frame, requestType);
        roll.text = CreateFontString(roll);
        roll.text:SetJustifyH("RIGHT");
        roll.text:SetPoint("RIGHT", roll, -10, 1);

        local notes = CreateElement(frame, roll);
        notes:SetPoint("TOPRIGHT", frame);
        notes.text = CreateFontString(notes);
        notes.text:ClearAllPoints();
        notes.text:SetPoint("TOPLEFT", notes, 0, 1);
        notes.text:SetPoint("BOTTOMRIGHT", notes, -2, 1);
        notes.text:SetWordWrap(true);
        notes:SetScript("OnEnter", function(self)
            if self.text:IsTruncated() then
                _G.ShowUIPanel(_G.GameTooltip);
                _G.GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT");
                _G.GameTooltip:SetText(self.text:GetText(), 1, 1, 1, 1, true);
                _G.GameTooltip:Show();
            end
            self:GetParent():RequestHighlight(true);
        end);
        notes:SetScript("OnLeave", function(self)
            _G.GameTooltip:Hide();
            self:GetParent():RequestHighlight(false);
        end);

        -- create widget
        local widget = {
            player = player,
            rank = rank,
            ep = ep,
            gp = gp,
            priority = priority,
            equipped = equipped,
            requestType = requestType,
            roll = roll,
            notes = notes,

            background = background,

            frame = frame,
            type  = Type
        }
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_Priority", 1;

    local mainSpecFont = CreateFont("ABGPHighlight");
    mainSpecFont:CopyFontObject(GameFontHighlight);
    mainSpecFont:SetTextColor(unpack(ABGP.ColorTable));

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self:SetHeight(20);

            self.frame.highlightRequests = 0;
            self.frame:UnlockHighlight();

            self.background:Hide();
        end,

        ["SetData"] = function(self, data)
            self.data = data;

            self.order.text:SetText(data.order or "");
            self.player.text:SetText(ABGP:ColorizeName(data.player or "", data.class));
            self.rank.text:SetText(data.rank or "");
            self.ep.text:SetText(data.ep and ("%.3f"):format(data.ep) or "--");
            self.gp.text:SetText(data.gp and ("%.3f"):format(data.gp) or "--");
            self.priority.text:SetText(data.priority and ("%.3f"):format(data.priority) or "--");

            local specialFont = data.important and "ABGPHighlight" or "GameFontNormal";
            self.order.text:SetFontObject(specialFont);
            self.player.text:SetFontObject(specialFont);
            self.rank.text:SetFontObject(specialFont);
            self.ep.text:SetFontObject(specialFont);
            self.gp.text:SetFontObject(specialFont);
            self.priority.text:SetFontObject(specialFont);
        end,

        ["SetWidths"] = function(self, widths)
            self.order:SetWidth(widths[1] or 0);
            self.player:SetWidth(widths[2] or 0);
            self.rank:SetWidth(widths[3] or 0);
            self.ep:SetWidth(widths[4] or 0);
            self.gp:SetWidth(widths[5] or 0);
            self.priority:SetWidth(widths[6] or 0);
        end,

        ["ShowBackground"] = function(self, show)
            self.background[show and "Show" or "Hide"](self.background);
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local frame = CreateFrame("Button");
        frame:SetHeight(32);
        frame:Hide();

        frame.highlightRequests = 0;
        frame.RequestHighlight = function(self, enable)
            self.highlightRequests = self.highlightRequests + (enable and 1 or -1);
            self[self.highlightRequests > 0 and "LockHighlight" or "UnlockHighlight"](self);
        end;

        local highlight = frame:CreateTexture(nil, "HIGHLIGHT");
        highlight:SetTexture("Interface\\HelpFrame\\HelpFrameButton-Highlight");
        highlight:SetAllPoints();
        highlight:SetBlendMode("ADD");
        highlight:SetTexCoord(0, 1, 0, 0.578125);

        local background = frame:CreateTexture(nil, "BACKGROUND");
        background:SetAllPoints();
        background:SetColorTexture(0, 0, 0, 0.5);

        local order = CreateElement(frame);
        order.text = CreateFontString(order);
        order.text:SetJustifyH("RIGHT");
        order.text:SetPoint("LEFT", order, 2, 1);
        order.text:SetPoint("RIGHT", order, -10, 1);

        local player = CreateElement(frame, order);
        player.text = CreateFontString(player);

        local rank = CreateElement(frame, player);
        rank.text = CreateFontString(rank);

        local ep = CreateElement(frame, rank);
        ep.text = CreateFontString(ep);

        local gp = CreateElement(frame, ep);
        gp.text = CreateFontString(gp);

        local priority = CreateElement(frame, gp);
        priority.text = CreateFontString(priority);

        -- create widget
        local widget = {
            order = order,
            player = player,
            rank = rank,
            ep = ep,
            gp = gp,
            priority = priority,

            background = background,

            frame = frame,
            type  = Type
        }
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_Item", 1;

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self.button:SetText("")
            self.itemLink.text:SetText("");

            self.frame.highlightRequests = 0;
            self.frame:UnlockHighlight();
        end,

        ["SetData"] = function(self, data)
            self.data = data;

            local fonts = {
                [ABGP.ItemRanks.HIGH] = "ABGPHighlight",
                [ABGP.ItemRanks.NORMAL] = "GameFontNormal",
                [ABGP.ItemRanks.LOW] = "GameFontDisable",
            };
            local font = fonts[ABGP:GetItemRank(data.itemLink)];
            self.itemLink.text:SetFontObject(font);

            self.itemLink.text:SetText(data.itemLink:match("(|H.+|h)"));
        end,

        ["SetText"] = function(self, text)
            self.button:SetText(text);
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local frame = CreateFrame("Button");
        frame:SetHeight(24);
        frame:Hide();

        frame.highlightRequests = 0;
        frame.RequestHighlight = function(self, enable)
            self.highlightRequests = self.highlightRequests + (enable and 1 or -1);
            self[self.highlightRequests > 0 and "LockHighlight" or "UnlockHighlight"](self);
        end;

        local highlight = frame:CreateTexture(nil, "HIGHLIGHT");
        highlight:SetTexture("Interface\\HelpFrame\\HelpFrameButton-Highlight");
        highlight:SetAllPoints();
        highlight:SetBlendMode("ADD");
        highlight:SetTexCoord(0, 1, 0, 0.578125);

        local button = CreateElement(frame, nil, "UIPanelButtonTemplate");
        if _G.ElvUI then
            _G.ElvUI[1]:GetModule("Skins"):HandleButton(button);
        end
        button:SetWidth(85);
        button:ClearAllPoints();
        button:SetPoint("TOPRIGHT", frame, -1, -1);
        button:SetPoint("BOTTOMRIGHT", frame, -1, 3);

        local itemLink = CreateElement(frame);
        itemLink.text = CreateFontString(itemLink);
        itemLink:ClearAllPoints();
        itemLink:SetPoint("TOPLEFT", frame);
        itemLink:SetPoint("BOTTOMLEFT", frame);
        itemLink:SetPoint("RIGHT", button, "LEFT");

        -- create widget
        local widget = {
            itemLink = itemLink,
            button = button,

            frame = frame,
            type  = Type
        }
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_ItemHistory", 1;

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self.player.text:SetText("");
            self.gp.text:SetText("");
            self.date.text:SetText("");
            self.itemLink.text:SetText("");

            self.frame.highlightRequests = 0;
            self.frame:UnlockHighlight();

            self.background:Hide();
        end,

        ["SetData"] = function(self, data)
            self.data = data;

            self.player.text:SetText(ABGP:ColorizeName(data.player or "", data.class));
            self.gp.text:SetText(data.gp);
            self.date.text:SetText(data.date);
            self.itemLink.text:SetText(data.itemLink or data.item);
        end,

        ["SetWidths"] = function(self, widths)
            self.player:SetWidth(widths[1] or 0);
            self.date:SetWidth(widths[2] or 0);
            self.gp:SetWidth(widths[3] or 0);
        end,

        ["ShowBackground"] = function(self, show)
            self.background[show and "Show" or "Hide"](self.background);
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local frame = CreateFrame("Button");
        frame:SetHeight(20);
        frame:Hide();

        frame.highlightRequests = 0;
        frame.RequestHighlight = function(self, enable)
            self.highlightRequests = self.highlightRequests + (enable and 1 or -1);
            self[self.highlightRequests > 0 and "LockHighlight" or "UnlockHighlight"](self);
        end;

        local highlight = frame:CreateTexture(nil, "HIGHLIGHT");
        highlight:SetTexture("Interface\\HelpFrame\\HelpFrameButton-Highlight");
        highlight:SetAllPoints();
        highlight:SetBlendMode("ADD");
        highlight:SetTexCoord(0, 1, 0, 0.578125);

        local background = frame:CreateTexture(nil, "BACKGROUND");
        background:SetAllPoints();
        background:SetColorTexture(0, 0, 0, 0.5);

        local player = CreateElement(frame);
        player.text = CreateFontString(player);

        local date = CreateElement(frame, player);
        date.text = CreateFontString(date);

        local gp = CreateElement(frame, date);
        gp.text = CreateFontString(gp);
        gp.text:SetJustifyH("RIGHT");
        gp.text:SetPoint("LEFT", gp, 2, 1);
        gp.text:SetPoint("RIGHT", gp, -10, 1);

        local itemLink = CreateElement(frame, gp);
        itemLink.text = CreateFontString(itemLink);
        itemLink:SetPoint("TOPRIGHT", frame);

        -- create widget
        local widget = {
            player = player,
            gp = gp,
            date = date,
            itemLink = itemLink,

            background = background,

            frame = frame,
            type  = Type
        }
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_ItemValue", 1;

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self.item.text:SetText("");
            self.gp.text:SetText("");
            self.notes.text:SetText("");
            self.priority.text:SetText("");

            self.frame.highlightRequests = 0;
            self.frame:UnlockHighlight();

            self.background:Hide();
        end,

        ["SetData"] = function(self, data)
            self.data = data;

            self.item.text:SetText(data[ABGP.ItemDataIndex.ITEMLINK] or data[ABGP.ItemDataIndex.NAME]);
            self.gp.text:SetText(data[ABGP.ItemDataIndex.GP]);
            self.notes.text:SetText(data[ABGP.ItemDataIndex.NOTES] and "[Note]" or "");
            self.priority.text:SetText(table.concat(data[ABGP.ItemDataIndex.PRIORITY], ", "));

            local font =ABGP:IsItemFavorited(data[ABGP.ItemDataIndex.ITEMLINK]) and "ABGPHighlight" or "GameFontNormal";
            self.gp.text:SetFontObject(font);
            self.notes.text:SetFontObject(font);
            self.priority.text:SetFontObject(font);
        end,

        ["SetWidths"] = function(self, widths)
            self.item:SetWidth(widths[1] or 0);
            self.gp:SetWidth(widths[2] or 0);
            self.notes:SetWidth(widths[3] or 0);
        end,

        ["ShowBackground"] = function(self, show)
            self.background[show and "Show" or "Hide"](self.background);
        end,

        ["EditPriorities"] = function(self)
            self.frame:SetHeight(30);
            if not self.priorityEditor then
                local priorityEditor = AceGUI:Create("ABGP_Filter");
                priorityEditor.frame:ClearAllPoints();
                priorityEditor.frame:SetPoint("TOPLEFT", self.notes, "TOPRIGHT");
                priorityEditor.frame:SetPoint("TOPRIGHT", self.frame);
                priorityEditor.frame:SetParent(self.frame);

                self.currentPriorities = {};
                for _, pri in ipairs(self.data[ABGP.ItemDataIndex.PRIORITY]) do self.currentPriorities[pri] = true; end

                priorityEditor:SetValues(self.currentPriorities, false, ABGP:GetItemPriorities());
                priorityEditor:SetCallback("OnClosed", function()
                    self.frame:SetHeight(20);
                    self.priority:Show();
                    self.priorityEditor.frame:Hide();

                    self.data[ABGP.ItemDataIndex.PRIORITY] = {};
                    for pri, value in pairs(self.currentPriorities) do
                        if value then table.insert(self.data[ABGP.ItemDataIndex.PRIORITY], pri); end
                    end
                    table.sort(self.data[ABGP.ItemDataIndex.PRIORITY]);

                    self:SetData(self.data);
                    self:Fire("OnPrioritiesUpdated");
                end);
                priorityEditor:SetText(table.concat(self.data[ABGP.ItemDataIndex.PRIORITY], ", "));
                self.priorityEditor = priorityEditor;
            end

            self.priority:Hide();
            self.priorityEditor.frame:Show();
            self.priorityEditor.button:Click();
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local frame = CreateFrame("Button");
        frame:SetHeight(20);
        frame:Hide();

        frame.highlightRequests = 0;
        frame.RequestHighlight = function(self, enable)
            self.highlightRequests = self.highlightRequests + (enable and 1 or -1);
            self[self.highlightRequests > 0 and "LockHighlight" or "UnlockHighlight"](self);
        end;

        local highlight = frame:CreateTexture(nil, "HIGHLIGHT");
        highlight:SetTexture("Interface\\HelpFrame\\HelpFrameButton-Highlight");
        highlight:SetAllPoints();
        highlight:SetBlendMode("ADD");
        highlight:SetTexCoord(0, 1, 0, 0.578125);

        local background = frame:CreateTexture(nil, "BACKGROUND");
        background:SetAllPoints();
        background:SetColorTexture(0, 0, 0, 0.5);

        local item = CreateElement(frame);
        item.text = CreateFontString(item);

        local gp = CreateElement(frame, item);
        gp.text = CreateFontString(gp);
        gp.text:SetJustifyH("RIGHT");
        gp.text:SetPoint("LEFT", gp, 2, 1);
        gp.text:SetPoint("RIGHT", gp, -10, 1);

        local notes = CreateElement(frame, gp);
        notes.text = CreateFontString(notes);
        notes:SetScript("OnEnter", function(self)
            local notes = self:GetParent().obj.data[ABGP.ItemDataIndex.NOTES];
            if notes then
                _G.ShowUIPanel(_G.GameTooltip);
                _G.GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT");
                _G.GameTooltip:SetText(notes, 1, 1, 1, 1, true);
                _G.GameTooltip:Show();
            end
            self:GetParent():RequestHighlight(true);
        end);
        notes:SetScript("OnLeave", function(self)
            _G.GameTooltip:Hide();
            self:GetParent():RequestHighlight(false);
        end);

        local priority = CreateElement(frame, notes);
        priority.text = CreateFontString(priority);
        priority:SetPoint("TOPRIGHT", frame);
        priority:SetScript("OnEnter", function(self)
            if self.text:IsTruncated() then
                _G.ShowUIPanel(_G.GameTooltip);
                _G.GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT");
                _G.GameTooltip:SetText(self.text:GetText(), 1, 1, 1, 1, true);
                _G.GameTooltip:Show();
            end
            self:GetParent():RequestHighlight(true);
        end);
        priority:SetScript("OnLeave", function(self)
            _G.GameTooltip:Hide();
            self:GetParent():RequestHighlight(false);
        end);

        -- create widget
        local widget = {
            item = item,
            gp = gp,
            notes = notes,
            priority = priority,

            background = background,

            frame = frame,
            type  = Type
        }
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_Paginator", 1;

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self.dataCount = 0;
            self.page = 1;
            self:CalculateRange(true);
        end,

        ["SetValues"] = function(self, dataCount, pageSize)
            self.dataCount = dataCount;
            self.pageSize = pageSize;
            self.pageCount = floor(self.dataCount / self.pageSize) + ((mod(self.dataCount, self.pageSize) == 0) and 0 or 1);
            self.page = min(max(1, self.page), self.pageCount);
            self:CalculateRange(true);
        end,

        ["CalculateRange"] = function(self, suppressEvent)
            if self.dataCount == 0 then
                self.first = 0;
                self.last = 0;

                self.firstBtn:SetDisabled(true);
                self.prevBtn:SetDisabled(true);

                self.nextBtn:SetDisabled(true);
                self.lastBtn:SetDisabled(true);

                self.text:SetText("");
            else
                self.first = 1 + (self.pageSize * (self.page - 1));
                self.last = min(self.first + self.pageSize - 1, self.dataCount);

                self.firstBtn:SetDisabled(self.page == 1);
                self.prevBtn:SetDisabled(self.page == 1);

                self.nextBtn:SetDisabled(self.page == self.pageCount);
                self.lastBtn:SetDisabled(self.page == self.pageCount);

                self.text:SetText(("Showing %d-%d of %d"):format(self.first, self.last, self.dataCount));
            end

            if not suppressEvent then
                self:Fire("OnRangeSet", self.first, self.last);
            end
        end,

        ["SetPage"] = function(self, page)
            self.page = page;
        end,

        ["GetRange"] = function(self)
            return self.first, self.last;
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local widget = {};

        local container = AceGUI:Create("SimpleGroup");
        container:SetFullWidth(true);
        container:SetLayout("table");
        container:SetUserData("table", { columns = { 0, 0, 1.0, 0, 0 } });

        local left1 = AceGUI:Create("Button");
        left1:SetWidth(45);
        left1:SetText("<<");
        left1:SetCallback("OnClick", function()
            widget.page = 1;
            widget:CalculateRange();
        end);
        container:AddChild(left1);

        local left2 = AceGUI:Create("Button");
        left2:SetWidth(40);
        left2:SetText("<");
        left2:SetCallback("OnClick", function()
            widget.page = widget.page - 1;
            widget:CalculateRange();
        end);
        container:AddChild(left2);

        local mid = AceGUI:Create("ABGP_Header");
        mid:SetFullWidth(true);
        mid:SetJustifyH("CENTER");
        container:AddChild(mid);

        local right1 = AceGUI:Create("Button");
        right1:SetWidth(40);
        right1:SetText(">");
        right1:SetCallback("OnClick", function()
            widget.page = widget.page + 1;
            widget:CalculateRange();
        end);
        container:AddChild(right1);

        local right2 = AceGUI:Create("Button");
        right2:SetWidth(45);
        right2:SetText(">>");
        right2:SetCallback("OnClick", function()
            widget.page = widget.pageCount;
            widget:CalculateRange();
        end);
        container:AddChild(right2);

        -- create widget
        widget.container = container;
        widget.firstBtn = left1;
        widget.prevBtn = left2;
        widget.text = mid;
        widget.nextBtn = right1;
        widget.lastBtn = right2;
        widget.frame = container.frame;
        widget.type  = Type;
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_Filter", 1;

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]

    local methods = {
        ["OnAcquire"] = function(self)
            self:DropdownOnAcquire();
            self:SetMultiselect(true);
        end,

        ["SetValues"] = function(self, allowed, showAllButton, values, sorted)
            self.allowed = allowed;
            self.values = values;
            self.showAllButton = showAllButton;

            if showAllButton then
                values.ALL = "All";
                table.insert(sorted, "ALL");
            end

            self:SetList(values, sorted);

            self:SetCallback("OnValueChanged", nil);
            self:UpdateCheckboxes();
            self:SetCallback("OnValueChanged", self.ValueChangedCallback);
        end,

        ["ValueChangedCallback"] = function(self, event, value, checked)
            if value == "ALL" then
                if checked then
                    for value in pairs(self.values) do
                        if value ~= "ALL" then self.allowed[value] = true; end
                    end
                end
            else
                if checked then
                    if self:ShowingAll() then
                        table.wipe(self.allowed);
                    end
                end
                self.allowed[value] = checked;
                if self:ShowingNone() and self.showAllButton then
                    for value in pairs(self.values) do
                        if value ~= "ALL" then self.allowed[value] = true; end
                    end
                end
            end

            self:SetCallback("OnValueChanged", nil);
            self:UpdateCheckboxes();
            self:SetCallback("OnValueChanged", self.ValueChangedCallback);
            self:Fire("OnFilterUpdated");
        end,

        ["UpdateCheckboxes"] = function(self)
            local all = self:ShowingAll();
            for value in pairs(self.values) do
                if value == "ALL" then
                    self:SetItemValue(value, all);
                else
                    self:SetItemValue(value, (not all or not self.showAllButton) and self.allowed[value]);
                end
            end
        end,

        ["ShowingAll"] = function(self)
            for value in pairs(self.values) do
                if value ~= "ALL" and not self.allowed[value] then return false; end
            end

            return true;
        end,

        ["ShowingNone"] = function(self)
            for _, state in pairs(self.allowed) do
                if state then return false; end
            end

            return true;
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local dropdown = AceGUI:Create("Dropdown");

        dropdown.type = Type;
        dropdown.DropdownOnAcquire = dropdown.OnAcquire;
        for method, func in pairs(methods) do
            dropdown[method] = func;
        end

        return dropdown;
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_Header", 1;

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self.text:SetText("");
            self.frame:SetWidth(100);
            self.frame:SetHeight(16);
            self.text:SetJustifyH("LEFT");
            self.text:SetJustifyV("CENTER");
            self.text:SetPoint("LEFT", self.frame, 2, 1);
            self.text:SetPoint("RIGHT", self.frame, -2, 1);
        end,

        ["SetText"] = function(self, text)
            self.text:SetText(text);
        end,

        ["SetWidth"] = function(self, width)
            self.frame:SetWidth(width);
        end,

        ["SetHeight"] = function(self, height)
            self.frame:SetHeight(height);
        end,

        ["SetJustifyH"] = function(self, justify)
            self.text:SetJustifyH(justify);
        end,

        ["SetJustifyV"] = function(self, justify)
            self.text:SetJustifyV(justify);
        end,

        ["SetPadding"] = function(self, left, right)
            self.text:SetPoint("LEFT", self.frame, left, 1);
            self.text:SetPoint("RIGHT", self.frame, right, 1);
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local frame = CreateFrame("Button");
        frame:SetHeight(16);
        frame:Hide();

        local text = CreateFontString(frame);
        text:SetFontObject(_G.GameFontHighlight);

        -- create widget
        local widget = {
            text = text,

            frame = frame,
            type  = Type
        }
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_AuditLog", 1;

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self.date.text:SetText("");
            self.auditType.text:SetText("");
            self.audit.text:SetText("");

            self.frame.highlightRequests = 0;
            self.frame:UnlockHighlight();

            self.background:Hide();
        end,

        ["SetData"] = function(self, data)
            self.data = data;

            self.date.text:SetText(data.date);
            self.auditType.text:SetText(data.type);
            self.audit.text:SetText(data.audit);

            local specialFont = (data.important) and "ABGPHighlight" or "GameFontHighlight";
            self.auditType.text:SetFontObject(specialFont);
            self.audit.text:SetFontObject(specialFont);
        end,

        ["SetWidths"] = function(self, widths)
            self.date:SetWidth(widths[1] or 0);
            self.auditType:SetWidth(widths[2] or 0);
        end,

        ["ShowBackground"] = function(self, show)
            self.background[show and "Show" or "Hide"](self.background);
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local frame = CreateFrame("Button");
        frame:SetHeight(20);
        frame:Hide();

        frame.highlightRequests = 0;
        frame.RequestHighlight = function(self, enable)
            self.highlightRequests = self.highlightRequests + (enable and 1 or -1);
            self[self.highlightRequests > 0 and "LockHighlight" or "UnlockHighlight"](self);
        end;

        local highlight = frame:CreateTexture(nil, "HIGHLIGHT");
        highlight:SetTexture("Interface\\HelpFrame\\HelpFrameButton-Highlight");
        highlight:SetAllPoints();
        highlight:SetBlendMode("ADD");
        highlight:SetTexCoord(0, 1, 0, 0.578125);

        local background = frame:CreateTexture(nil, "BACKGROUND");
        background:SetAllPoints();
        background:SetColorTexture(0, 0, 0, 0.5);

        local date = CreateElement(frame);
        date.text = CreateFontString(date);

        local auditType = CreateElement(frame, date);
        auditType.text = CreateFontString(auditType);

        local audit = CreateElement(frame, auditType);
        audit.text = CreateFontString(audit);
        audit:SetPoint("TOPRIGHT", frame);
        audit:SetScript("OnEnter", function(self)
            if self.text:IsTruncated() then
                _G.ShowUIPanel(_G.GameTooltip);
                _G.GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT");
                _G.GameTooltip:SetText(self.text:GetText(), 1, 1, 1, 1, true);
                _G.GameTooltip:Show();
            end
            self:GetParent():RequestHighlight(true);
        end);
        audit:SetScript("OnLeave", function(self)
            _G.GameTooltip:Hide();
            self:GetParent():RequestHighlight(false);
        end);

        -- create widget
        local widget = {
            date = date,
            auditType = auditType,
            audit = audit,

            background = background,

            frame = frame,
            type  = Type
        }
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end

do
    local Type, Version = "ABGP_LootFrame", 1;

    local function Frame_OnEvent(frame, event)
        local self = frame.obj;
        self:SetItem(self:GetItem());
    end

    local function Frame_OnUpdate(frame, elapsed)
        local self = frame.obj;
        self.elapsed = self.elapsed + elapsed;
        if self.duration and (MouseIsOver(frame) or ABGP:IsContextMenuOpen()) then
            self.elapsed = math.min(self.elapsed, self.duration - 1);
        end

        if self.elapsed <= self.fadeIn then
            -- Fade in
            local alpha = self.elapsed / self.fadeIn;
            frame:SetAlpha(alpha);
        elseif not self.duration or self.elapsed <= self.duration then
            -- Normal state
            frame:SetAlpha(1);
        elseif self.elapsed >= self.duration + self.fadeOut then
            -- Completed
            frame:Hide();
        else
            -- Fade out
            local alpha = 1 - (self.elapsed - self.duration) / self.fadeOut;
            frame:SetAlpha(alpha);
        end
    end

    local function Frame_OnMouseDown(frame, button)
        if button == "LeftButton" then
            frame.obj:Fire("OnMouseDown");
        end
    end

    local function Frame_OnMouseUp(frame, button)
        if button == "LeftButton" then
            frame.obj:Fire("OnMouseUp");
        elseif button == "RightButton" then
            frame.obj:Fire("OnClick", button, false);
        end
    end

    local function Frame_OnHide(frame)
        frame.obj:Fire("OnHide");
    end

    local function Button_OnClick(frame, button, down)
        local self = frame:GetParent().obj;
        local itemLink = self.itemLink;
        if not itemLink then return; end

        if IsModifiedClick() then
            _G.HandleModifiedItemClick(itemLink);
        else
            self:Fire("OnClick", button, down);
        end
    end

    local frameCount = 0;

    --[[-----------------------------------------------------------------------------
    Methods
    -------------------------------------------------------------------------------]]
    local methods = {
        ["OnAcquire"] = function(self)
            self.frame:SetAlpha(0);
            self.frame:Show();

            self.itemLink = nil;
            self.elapsed = 0;
            self.fadeIn = 0.2;
            self.duration = nil;
            self.fadeOut = nil;
        end,

        ["OnRelease"] = function(self)
            self.frame:UnregisterEvent("GET_ITEM_INFO_RECEIVED");
        end,

        ["GetItem"] = function(self)
            return self.itemLink;
        end,

        ["SetItem"] = function(self, itemLink)
            self.itemLink = itemLink;

            local frame = self.frame;
            local itemName = ABGP:GetItemName(itemLink);
            local value = ABGP:GetItemValue(itemName);
            local _, _, rarity = GetItemInfo(itemLink);

            if rarity then
                frame:UnregisterEvent("GET_ITEM_INFO_RECEIVED");
            else
                rarity = LE_ITEM_QUALITY_COMMON;
                frame:RegisterEvent("GET_ITEM_INFO_RECEIVED");
            end
            local r, g, b = GetItemQualityColor(rarity);

            if frame.elvui then
                if ABGP:IsItemFavorited(itemLink) then
                    rarity = LE_ITEM_QUALITY_ARTIFACT;
                    r, g, b = GetItemQualityColor(rarity);
                end

                frame.button.icon:SetTexture(GetItemIcon(itemLink));
                frame.button.link = itemLink;

                frame.fsloot:SetText(itemName);

                frame.status:SetStatusBarColor(r, g, b, .7);
                frame.status.bg:SetColorTexture(r, g, b);

                local color = ABGP.ColorTable;
                local valueText = (value and value.gp ~= 0) and value.gp or "--";
                frame.fsbind:SetText(valueText);
                frame.fsbind:SetVertexColor(color.r, color.g, color.b);
            else
                frame.IconFrame.Icon:SetTexture(GetItemIcon(itemLink));

                frame.Name:SetVertexColor(r, g, b);
                frame.Name:SetText(itemName);

                local valueText = (value and value.gp ~= 0) and ("GP cost: %s%d|r"):format(ABGP.Color, value.gp) or "No GP Cost";
                frame.Cost:SetVertexColor(1, 1, 1);
                frame.Cost:SetText(valueText);

                if ABGP:IsItemFavorited(itemLink) then
                    frame:SetBackdrop({
                        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Background",
                        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
                        tile = true,
                        tileSize = 32,
                        edgeSize = 32,
                        insets = { left = 11, right = 12, top = 12, bottom = 11 }
                    });
                    _G[frame:GetName().."Corner"]:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Gold-Corner");
                    _G[frame:GetName().."Decoration"]:Show();
                else
                    frame:SetBackdrop({
                        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
                        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
                        tile = true,
                        tileSize = 32,
                        edgeSize = 32,
                        insets = { left = 11, right = 12, top = 12, bottom = 11 }
                    });
                    _G[frame:GetName().."Corner"]:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Corner");
                    _G[frame:GetName().."Decoration"]:Hide();
                end
            end
        end,

        ["SetDuration"] = function(self, duration, fadeOut)
            self.elapsed = 0;
            self.duration = duration;
            self.fadeOut = fadeOut or 1;
        end,
    }

    --[[-----------------------------------------------------------------------------
    Constructor
    -------------------------------------------------------------------------------]]
    local function Constructor()
        local frame, button;
        if ABGP:Get("lootElvUI") and _G.ElvUI and _G.ElvUI[1].private.general.lootRoll then
            frame = _G.ElvUI[1]:GetModule("Misc"):CreateRollFrame();
            frame.elvui = true;

            -- Change default width
            frame:SetWidth(260);
            frame.fsbind:SetWidth(30);

            -- "Hide" need/greed/pass buttons
            frame.needbutt:SetAlpha(0);
            frame.needbutt:EnableMouse(false);
            frame.needbutt:SetWidth(1);
            frame.greedbutt:SetAlpha(0);
            frame.greedbutt:EnableMouse(false);
            frame.greedbutt:SetWidth(1);
            frame.pass:GetParent():SetAlpha(0);
            frame.pass:GetParent():EnableMouse(false);
            frame.pass:GetParent():SetWidth(1);

            button = frame.button;
        else
            frameCount = frameCount + 1;
            frame = CreateFrame("Frame", "ABGP_LootFrame" .. frameCount, nil, "ABGPLootTemplate");
            button = frame.IconFrame;
        end

        frame:SetScript("OnEvent", Frame_OnEvent);
        frame:SetScript("OnUpdate", Frame_OnUpdate);
        frame:SetScript("OnHide", Frame_OnHide);

        frame:EnableMouse(true);
        frame:SetClampedToScreen(true);
        frame:SetScript("OnMouseDown", Frame_OnMouseDown);
        frame:SetScript("OnMouseUp", Frame_OnMouseUp);

        button.hasItem = true;
        button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
        button:SetScript("OnClick", Button_OnClick);

        -- create widget
        local widget = {
            frame = frame,
            type  = Type
        }
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(Type, Constructor, Version)
end
